#include "sudoku.hpp"
#include <fstream>
#include <iostream>
#include <string>


Sudoku::Sudoku(void) {
    board.resize(9, std::vector<int>(9, -1));
}

void Sudoku::FillBoard(std::string filePath) {
    std::ifstream sudokuFile(filePath);
    if (!sudokuFile) {
        std::cerr << "Error opening file: " << filePath << std::endl;
        exit(1);
    }
    std::string line;
    int height = 0;
    while (!sudokuFile.eof() && height < 9) { 
        sudokuFile >> line;
        for (int i = 0; i < 9; ++i) { board[height][i] = line[i] - '0'; }
        ++height;
    }
}

// Base global call for solving
void Sudoku::Solve(void) {
    if (!_Solve()) { std::cerr << "No solution found" << std::endl; }
}

bool Sudoku::IsBoardSolved(void) {
    int row, column;
    for (row = 0; row < 9; row += 3) {
        for (column = 0; column < 9; column += 3) {
            if (!GetUnusedRow(row).empty()) { return false; }
            if (!GetUnusedColumn(column).empty()) { return false; }
            if (!GetUnusedSubgrid(row, column).empty()) { return false; }
        }
    }
    return true;
}

void Sudoku::Print(void) {
    std::cout << "===================" << std::endl;
    for (auto i : board) {
        std::cout << '|';
        for (auto j : i) {
            std::cout << j << '|';
        }
        std::cout << std::endl;
    }
    std::cout << "===================" << std::endl;
}

void Sudoku::RawPrint(void) {
    for (auto i : board) {
        for (auto j : i) {
            std::cout << j;
        }
        std::cout << std::endl;
    }
}

// Recursive call for solving
bool Sudoku::_Solve(void) {
    int row, column;
    if (!FindEmptyCell(row, column)) {
        if (IsBoardSolved()) { return true; }
        return false;
    }
    std::vector<int> rowPossiblities, columnPossibilities, subgridPossibilities;
    rowPossiblities = GetUnusedRow(row);
    columnPossibilities = GetUnusedColumn(column);
    subgridPossibilities = GetUnusedSubgrid(row, column);
    for (int i : rowPossiblities) {
        for (int j : columnPossibilities) {
            for (int k : subgridPossibilities) {
                if (i == j && j == k) { 
                    board[row][column] = i; 
                    if (_Solve()) { return true; }
                    board[row][column] = 0;
                }
            }
        }
    }
    return false;
}

// Fills column and row of an empty cell and returns true
// If no empty cell is found, return false
bool Sudoku::FindEmptyCell(int& row, int& column) {
    for (row = 0; row < 9; ++(row)) {
        for (column = 0; column < 9; ++(column)) {
            if (board[row][column] == 0) { return true; }
        }
    }
    return false;
}

// Returns vector of unused numbers in row
std::vector<int> Sudoku::GetUnusedRow(const int& row) {
    std::vector<int> leftOver{0,1,2,3,4,5,6,7,8,9};
    int* box;
    for (int i = 0; i < 9; i++) {
        box = &board[row][i];
        if (leftOver[*box] == *box) { leftOver[*box] = 0; }
    }
    TrimPossiblities(leftOver);
    return leftOver;
}

// Returns vector of unused numbers in column
std::vector<int> Sudoku::GetUnusedColumn(const int& column) {
    std::vector<int> leftOver{0,1,2,3,4,5,6,7,8,9};
    int* box;
    for (int i = 0; i < 9; i++) {
        box = &board[i][column];
        if (leftOver[*box] == *box) { leftOver[*box] = 0; }
    }
    TrimPossiblities(leftOver);
    return leftOver;
}

// Returns vector of unused numbers in subgrid
std::vector<int> Sudoku::GetUnusedSubgrid(const int& row, const int& column) {
    std::vector<int> leftOver{0,1,2,3,4,5,6,7,8,9};
    int* box;
    int startRow = row - (row % 3);
    int startColumn = column - (column % 3);
    for (int i = 0; i < 3; ++i) {
        for (int j = 0; j < 3; ++j) {
            box = &board[startRow + i][startColumn + j]; 
            if (leftOver[*box] == *box) { leftOver[*box] = 0; }
        }
    }
    TrimPossiblities(leftOver);
    return leftOver;
}

void TrimPossiblities(std::vector<int>& possibilities) {
    for (auto it = possibilities.begin(); it != possibilities.end();) {
        if (*it == 0) { it = possibilities.erase(it); }
        else { ++it; }
    }
}
